/**********************************************************************
 * key_value.h                                            December 1999
 * Horms                                             horms@vergenet.net
 *
 * Key value pair
 *
 * Released under the terms of the GNU LGPL
 *
 **********************************************************************/

#ifndef KEY_VALUE_FRUB
#define KEY_VALUE_FRUB

#include <stdlib.h>

typedef struct {
  void *key;
  void (*destroy_key)(void *); 
  void *(*dup_key)(void *);
  void *value;
  void (*destroy_value)(void *);
  void *(*dup_value)(void *);
} key_value_t;


/**********************************************************************
 * key_value_create
 * Create a new key_value structure
 * pre: none
 * post: a new key_value strcture is allocated and elements are initialised
 * return: pointer to key_value_t
 *         NULL on error
 **********************************************************************/

key_value_t *key_value_create(void);


/**********************************************************************
 * key_value_destroy
 * Destory a key_value and its contents
 * pre: kv: key_value structure to destroy
 * post: kv and its contents are freed
 * return: none
 **********************************************************************/

void key_value_destroy(key_value_t *kv);


/**********************************************************************
 * key_value_duplicate
 * Make a copy of a key_value_structure
 * pre: kv: key_value structure to duplicate
 * post: 
 **********************************************************************/

key_value_t *key_value_duplicate(key_value_t *kv);


/**********************************************************************
 * key_value_assign
 * Assign a key and value to a key_value structure
 * pre: kv: key_valye_structure to assign key and value to
 *          key: key to assign
 *          key_detstory: pointer to a function to destroy key
 *                        Use NULL if you do not want the key destroyed
 *                        when the key_value_destroy is called
 *          key_duplicate: pointer to a function to duplicate key
 *                         Use NULL if you want the key duplicated
 *                         using new_key=old_key
 *          value: value to assign
 *          value_detstory: pointer to a function to destroy value
 *                        Use NULL if you do not want the value destroyed
 *                        when the value_value_destroy is called
 *          value_duplicate: pointer to a function to duplicate value
 *                         Use NULL if you want the value duplicated
 *                         using new_value=old_value
 * post: key and value are assigned and destiry functions are registered
 * return: pointer to key_value structure
 *         NULL if kv is NULL
 **********************************************************************/

key_value_t *key_value_assign(
  key_value_t *kv,
  void *key,
  void (*destroy_key)(void *), 
  void *(*dup_key)(void *), 
  void *value,
  void (*destroy_value)(void *),
  void *(*dup_value)(void *)
);


/**********************************************************************
 * key_value_unassign
 * Unassign valyes in key_values structure
 * Useful if you want to destroy the key_value structure
 * without freeing the contents
 * pre: kv: key_valye structure to unassign values of
 * post: All elements in key_value structure are set to NULL
 * return: pointer to key_value structure
 *         NULL if kv is NULL
 **********************************************************************/

key_value_t *key_value_unasign(key_value_t *kv);


/**********************************************************************
 * key_value_key_get_key
 * Get the key from a key_value structure
 * pre: kv: key value structure to get the key of
 * return: key of kv
 *         NULL if kv is NULL
 **********************************************************************/

void *key_value_get_key(key_value_t *kv);


/**********************************************************************
 * key_value_key_get_value
 * Get the value from a key_value structure
 * pre: kv: key value structure to get the value of
 * return: value of kv
 *         NULL if kv is NULL
 **********************************************************************/

void *key_value_get_value(key_value_t *kv);

#endif
