/**********************************************************************
 * imap4_write .c                                        September 1999
 * Horms                                             horms@vergenet.net
 *
 * Subroutines to write IMAP4 protocol output
 *
 * Licenced under the terms of the GNU GPL
 *
 **********************************************************************/

#include "imap4_write.h"
#include "options.h"


/**********************************************************************
 * imap4_write
 * Display an message of the form [<tag> <type> ]<string>
 * Pre: fd: file descriptor to write to
 *      flag: flag to pass to write_str as per str.h
 *      tag: tag to display
 *           if NULL, then IMAP4_UNTAGED is used
 *      type: type of message, IMAP4_OK, IMAP4_NO or IMAP4_BAD
 *            if NULL then only string is written, no tag and no type
 *      string: mesage to display
 * Return 0 on success
 *        -1 otherwise
 **********************************************************************/

int imap4_write(
  const int fd, 
  const flag_t flag,
  const token_t *tag, 
  const char *type, 
  const char *string
){
  char *tag_string;
  int free_tag_string=0;

  extern options_t opt;

  if(type==NULL){
    if(write_str(fd, flag, 1, string)<0){
      if(opt.debug){
        syslog(LOG_WARNING, "imap4_write: write_strings");
      }
      return(-1);
    }
  }
  else {
    if(tag==NULL){
      tag_string=IMAP4_UNTAGED;
    }
    else {
      if((tag_string=token_to_string(tag))==NULL){
        if(opt.debug){
          syslog(LOG_WARNING, "imap4_in_tagged_ok: token_to_string");
        }
        return(-1);
      }
      free_tag_string=1;
    }
    if(write_str(fd, flag, 5, tag_string, " ", type, " ", string)<0){
      if(opt.debug){
        syslog(LOG_WARNING, "imap4_write: write_strings");
      }
      return(-1);
    }
  }
  
  if(free_tag_string){
    str_free(&tag_string);
  }
  return(0);
}
